from pweb_ssr.lib.jinja_tag_helper import StandaloneTag
from pweb_ssr.table.table_common import TableCommon

table_common = TableCommon()


class SortableHeader(StandaloneTag):
    tags = {"sortable_header"}

    def render(self, name, display_name=None, *args, **kwargs):
        return table_common.sortable_header(name=name, display_name=display_name, kwargs=kwargs)


class Pagination(StandaloneTag):
    tags = {"pagination"}

    def render(self, current_page: int, total_page: int, *args, **kwargs):
        return table_common.pagination(current_page=current_page, total_page=total_page, kwargs=kwargs)


class SearchNameValue(StandaloneTag):
    tags = {"search_name_value"}

    def render(self, *args, **kwargs):
        return table_common.search_name_value()



from pweb_ssr.common.pweb_ssr_structure import SSRTemplateAssets


class PWebSSRConfig:
    DEFAULT_TEMPLATE_ASSETS: SSRTemplateAssets = None
    SSR_HTML_PATH: str = None

    # Form Input HTML Class Name
    INPUT_LABEL_CLASS_NAME = "form-label"
    INPUT_LABEL_CHECKBOX_CLASS_NAME = "form-check-label"
    INPUT_REQUIRED_SIGN_CLASS_NAME = "text-danger me-1"
    INPUT_CLASS_NAME = "form-control"
    CHECKBOX_CLASS_NAME = "form-check-input"
    CHECKBOX_WRAPPER_CLASS_NAME = "form-check"
    SELECT_CLASS_NAME = "form-select"
    INPUT_ERROR_CLASS_NAME = "is-invalid"
    INPUT_ERROR_MESSAGE_CLASS_NAME = "invalid-feedback"
    INPUT_HELP_MESSAGE_CLASS_NAME = "form-text"

    # Table Related Class
    TABLE_SORT_ICON_NAME = "fa-sort"
    TABLE_SORT_ASC_ICON_NAME = "fa-sort-down"
    TABLE_SORT_DESC_ICON_NAME = "fa-sort-up"
    TABLE_SORT_ICON_OTHER_CLASS_NAME = "me-1 fas"
    TABLE_SORT_LINK_CLASS_NAME = "text-decoration-none"
    TABLE_ITEM_PER_PAGE_OPTIONS: str = [25, 50, 100, 200, 500]



from pweb_ssr.common.pweb_ssr_init import PWebSSRInit


class PWebJinjaExtend:
    _global_variables = {
        "pweb_app_name": PWebSSRInit.get_app_name(),
    }

    _extensions = [
        "pweb_ssr.form.form_tag_registry.FormInput",
        "pweb_ssr.form.form_tag_registry.InputError",
        "pweb_ssr.form.form_tag_registry.ErrorClass",
        "pweb_ssr.form.form_tag_registry.InputHelp",
        "pweb_ssr.form.form_tag_registry.MakeChecked",
        "pweb_ssr.form.form_tag_registry.MakeSelect",
        "pweb_ssr.form.form_tag_registry.SetValue",

        "pweb_ssr.table.table_tag_registry.SortableHeader",
        "pweb_ssr.table.table_tag_registry.Pagination",
        "pweb_ssr.table.table_tag_registry.SearchNameValue",
    ]

    def register(self, pweb_app):
        self.register_extension(pweb_app)
        self.register_global_variable(pweb_app)

    def register_extension(self, pweb_app):
        for extension in self._extensions:
            pweb_app.jinja_env.add_extension(extension)

    def register_global_variable(self, pweb_app):
        if pweb_app and pweb_app.jinja_env and pweb_app.jinja_env.globals:
            for method in self._global_variables:
                if method not in pweb_app.jinja_env.globals:
                    pweb_app.jinja_env.globals[method] = self._global_variables[method]



from ppy_file_text import StringUtil
from pweb_form_rest import PWebFRConfig
from pweb_form_rest.crud.pweb_request_data import RequestData
from pweb_form_rest.ui.pweb_ui_helper import ssr_ui_render_html_file
from pweb_ssr.common.pweb_ssr_config import PWebSSRConfig
from pweb_ssr.common.pweb_ssr_html_file import PWebSSRHTMLFile


class TableCommon:
    html_file: PWebSSRHTMLFile = PWebSSRHTMLFile()
    request_data: RequestData = RequestData()

    def sortable_header(self, name, display_name, **kwargs):
        if not display_name:
            display_name = StringUtil.human_readable(name)

        icon = PWebSSRConfig.TABLE_SORT_ICON_NAME
        sort_field_key = PWebFRConfig.SORT_FIELD_PARAM_NAME
        sort_order_key = PWebFRConfig.SORT_ORDER_PARAM_NAME
        url_sort_field = self.request_data.get_query_args_value(sort_field_key, None)
        url_sort_order = self.request_data.get_query_args_value(sort_order_key, None)
        sort_fields = {sort_field_key: name, sort_order_key: "asc"}
        if url_sort_field and url_sort_field == name:
            icon = PWebSSRConfig.TABLE_SORT_ASC_ICON_NAME
            order = "asc"
            if url_sort_order == "asc":
                icon = PWebSSRConfig.TABLE_SORT_DESC_ICON_NAME
                order = "desc"
            sort_fields[sort_order_key] = order

        url_info = self.request_data.get_url_info()
        url = self.request_data.add_to_query_params(url_info.relativeURLWithParam, sort_fields)
        params = {
            "display_name": display_name,
            "icon": icon,
            "url": url,
            "icon_class": PWebSSRConfig.TABLE_SORT_ICON_OTHER_CLASS_NAME,
            "link_class": PWebSSRConfig.TABLE_SORT_LINK_CLASS_NAME
        }
        return ssr_ui_render_html_file(self.html_file.sortable_header_html(), params=params)

    def _page_number_calculate(self, current_page: int, total_page: int):
        delta = 2
        left = current_page - delta
        right = current_page + delta + 1
        range_list = []
        range_with_dots = []
        temp = 0

        for i in range(1, total_page + 1):
            if i == 1 or i == total_page or left <= i < right:
                range_list.append(i)

        for i in range_list:
            if temp:
                if (i - temp) == 2:
                    range_with_dots.append(temp + 1)
                elif (i - temp) != 1:
                    range_with_dots.append("...")
            range_with_dots.append(i)
            temp = i
        return range_with_dots

    def _prepare_pagination_link(self, current_page: int, total_page: int):
        pagination_item = self._page_number_calculate(current_page, total_page)
        pagination_details = []
        url_info = self.request_data.get_url_info()
        for pagination in pagination_item:
            selected = False
            url = "#"
            if pagination == current_page:
                selected = True
            elif pagination != "...":
                url = self.request_data.add_to_query_params(url_info.relativeURLWithParam, {PWebFRConfig.PAGE_PARAM_NAME: pagination})
            pagination_details.append({
                "text": pagination,
                "url": url,
                "selected": selected,
            })
        return pagination_details

    def pagination(self, current_page: int, total_page: int, **kwargs):
        if total_page <= 1:
            return ""
        params = {
            "prev": "",
            "pages": self._prepare_pagination_link(current_page, total_page),
            "next": "",
            "item_per_page_name": PWebFRConfig.ITEM_PER_PAGE_PARAM_NAME,
            "item_per_page_options": PWebSSRConfig.TABLE_ITEM_PER_PAGE_OPTIONS,
            "item_per_page_selected": self.request_data.get_query_args_value(PWebFRConfig.ITEM_PER_PAGE_PARAM_NAME, PWebFRConfig.TOTAL_ITEM_PER_PAGE),
        }
        return ssr_ui_render_html_file(self.html_file.pagination_html(), params=params)

    def search_name_value(self):
        value = self.request_data.get_query_args_value(PWebFRConfig.SEARCH_FIELD_PARAM_NAME, "#")
        if not value or value == "#":
            value = ""
        return "name='" + PWebFRConfig.SEARCH_FIELD_PARAM_NAME + "' value='" + value + "'"



from flask import Blueprint
import pweb_ssr.common.pweb_ssr_config
from application.config.app_config import Config
from ppy_common import ObjectHelper
from pweb_ssr import SSRTemplateAssets
from pweb_ssr.common.pweb_ssr_config import PWebSSRConfig


class PWebSSRInit:

    @staticmethod
    def init_default_template_assets(pweb_app):
        if PWebSSRConfig.DEFAULT_TEMPLATE_ASSETS and isinstance(PWebSSRConfig.DEFAULT_TEMPLATE_ASSETS, SSRTemplateAssets):
            ssr_template_assets = PWebSSRConfig.DEFAULT_TEMPLATE_ASSETS
            pweb_app.register_blueprint(Blueprint(
                name=ssr_template_assets.name,
                import_name=ssr_template_assets.import_name,
                static_folder=ssr_template_assets.static_folder,
                static_url_path=ssr_template_assets.static_url_path,
                template_folder=ssr_template_assets.template_folder,
                url_prefix=ssr_template_assets.url_prefix,
            ))

    @staticmethod
    def merge_config(config):
        ObjectHelper.copy_config_property(config, pweb_ssr.common.pweb_ssr_config.PWebSSRConfig)

    @staticmethod
    def get_app_name():
        return Config.APP_NAME



from pweb import PWebComponentRegister, PWebModuleDetails
from pweb_ssr.common.pweb_ssr_init import PWebSSRInit
from pweb_ssr.pweb_jinja_extend import PWebJinjaExtend


class PWebSSRModule(PWebComponentRegister):

    def app_details(self) -> PWebModuleDetails:
        return PWebModuleDetails(system_name="pweb-ssr", display_name="PWeb SSR Module")

    def run_on_cli_init(self, pweb_app, config):
        PWebSSRInit.merge_config(config=config)

    def register_model(self, pweb_db) -> list:
        pass

    def register_controller(self, pweb_app):
        pass

    def run_on_start(self, pweb_app, config):
        PWebSSRInit.merge_config(config=config)
        PWebSSRInit.init_default_template_assets(pweb_app)
        PWebJinjaExtend().register(pweb_app)



from pweb_form_rest import FormField
from pweb_ssr.form.from_input_common import FormInputCommon
from pweb_ssr.lib.jinja_tag_helper import StandaloneTag


form_input_common = FormInputCommon()


class FormInput(StandaloneTag):
    tags = {"form_input"}

    def render(self, field: FormField, *args, **kwargs):
        return form_input_common.get_form_input(field=field, kwargs=kwargs)


class InputError(StandaloneTag):
    tags = {"input_error"}

    def render(self, field: FormField, *args, **kwargs):
        return form_input_common.get_input_error(field=field, kwargs=kwargs)


class ErrorClass(StandaloneTag):
    tags = {"error_class"}

    def render(self, field: FormField, *args, **kwargs):
        return form_input_common.get_error_class(field=field, kwargs=kwargs)


class InputHelp(StandaloneTag):
    tags = {"input_help"}

    def render(self, field: FormField, *args, **kwargs):
        return form_input_common.get_input_help(field=field, kwargs=kwargs)


class MakeChecked(StandaloneTag):
    tags = {"make_checked"}

    def render(self, key, value, *args, **kwargs):
        return form_input_common.get_make_checked(key=key, value=value, kwargs=kwargs)


class MakeSelect(StandaloneTag):
    tags = {"make_select"}

    def render(self, key, value, *args, **kwargs):
        return form_input_common.get_make_select(key=key, value=value, kwargs=kwargs)


class SetValue(StandaloneTag):
    tags = {"set_value"}

    def render(self, value, *args, **kwargs):
        return form_input_common.get_set_value(value=value, kwargs=kwargs)



from typing import Union


class SSRTemplateAssets:
    name: str = None
    import_name: str = None
    static_folder: str = None
    static_url_path: str = None
    template_folder: str = None
    url_prefix: str = "/"

    def __init__(
            self,
            name: str,
            import_name: str,
            static_folder: str,
            static_url_path: str,
            template_folder: str,
            url_prefix: Union[str, None] = None,
    ):
        self.name = name
        self.import_name = import_name
        self.static_folder = static_folder
        self.static_url_path = static_url_path
        self.template_folder = template_folder
        if url_prefix:
            self.url_prefix = url_prefix



from copy import copy
from pweb_form_rest import FormField
from pweb_form_rest.ui.pweb_ui_helper import ssr_ui_render_html_file
from pweb_ssr.common.pweb_ssr_config import PWebSSRConfig
from pweb_ssr.common.pweb_ssr_html_file import PWebSSRHTMLFile


class FormInputCommon:
    html_file: PWebSSRHTMLFile = PWebSSRHTMLFile()

    def dict_to_attribute(self, dictionary: dict, ignore: list = []):
        attributes = ""
        for key in dictionary:
            if key in ignore:
                continue

            if dictionary[key] == True and key != "value":
                attributes += f" {key}"
            elif dictionary[key]:
                attributes += f" {key}='{dictionary[key]}'"
            attributes = attributes.strip()
        return attributes

    def get_kwargs_value(self, kwargs, key, concat=None, default: any = ""):
        value = None
        if key in kwargs:
            value = kwargs[key]
        if value and concat:
            return f"{value} {concat}".strip()
        elif value:
            return f"{value}".strip()
        elif concat:
            return f"{concat}".strip()
        return default

    def _get_wrapper_attribute(self, field: FormField, kwargs):
        wrapper_attribute_dict = {}
        concat = None
        if field.inputType == "checkbox" or field.inputType == "radio":
            concat = PWebSSRConfig.CHECKBOX_WRAPPER_CLASS_NAME
        wrapper_class = self.get_kwargs_value(kwargs=kwargs, key="wrapper", default=None, concat=concat)
        if wrapper_class:
            wrapper_attribute_dict["class"] = wrapper_class

        wrapper_id = self.get_kwargs_value(kwargs=kwargs, key="wrapper_id", default=None)
        if wrapper_id:
            wrapper_attribute_dict["id"] = wrapper_id

        return self.dict_to_attribute(dictionary=wrapper_attribute_dict)

    def _get_input_attribute(self, field: FormField, kwargs):
        input_class = PWebSSRConfig.INPUT_CLASS_NAME
        if field.inputType == "select":
            input_class = PWebSSRConfig.SELECT_CLASS_NAME
        elif field.inputType == "checkbox" or field.inputType == "radio":
            input_class = PWebSSRConfig.CHECKBOX_CLASS_NAME

        field.add_attribute("class", self.get_kwargs_value(kwargs=kwargs, key="input_class", concat=input_class))
        field.add_attribute("id", self.get_kwargs_value(kwargs=kwargs, key="input_id"))
        field.add_attribute("name", field.name)

        if field.isError:
            field.add_attribute("class", PWebSSRConfig.INPUT_ERROR_CLASS_NAME)

        if field.placeholder and field.inputType != "select":
            field.add_attribute("placeholder", field.placeholder)

        if field.required:
            field.add_attribute("required", True)

        if field.inputType != "textarea" and field.inputType != "select":
            field.add_attribute("value", field.value)
            field.add_attribute("type", field.inputType)

        return self.dict_to_attribute(field.allAttributes)

    def _get_select_options(self, field: FormField, params):
        if field.inputType != "select" or not field.selectOptions:
            return params

        options = []
        for item in field.selectOptions:
            label = field.selectOptions[item]
            option = {
                "label": label,
                "value": item,
            }
            if field.value and str(field.value) == str(item):
                option['selected'] = True
            elif field.defaultValue and str(field.defaultValue) == str(item):
                option['selected'] = True
            options.append(option)
        params["options"] = options
        return params

    def _set_radio_checkbox_attrs(self, field: FormField):
        if field.inputType != "checkbox" and field.inputType != "radio":
            return field

        if str(field.checked) == str(field.value):
            field.add_attribute("checked", True)
        field.value = f"{field.checked}"
        return field

    def _overwrite_by_kwargs(self, field: FormField, kwargs):
        overwrite_attrs = ["label", "value", "checked", "unchecked"]
        for name in overwrite_attrs:
            if name in kwargs and hasattr(field, name):
                setattr(field, name, kwargs[name])
        return field

    def get_form_input(self, field: FormField, kwargs):
        field = copy(field)
        field = self._overwrite_by_kwargs(field=field, kwargs=kwargs)
        field = self._set_radio_checkbox_attrs(field=field)
        params = {
            "label_required_class": PWebSSRConfig.INPUT_REQUIRED_SIGN_CLASS_NAME,
            "help_message_class": PWebSSRConfig.INPUT_HELP_MESSAGE_CLASS_NAME,
            "error_message_class": PWebSSRConfig.INPUT_ERROR_MESSAGE_CLASS_NAME,
            "wrapper_attributes": self._get_wrapper_attribute(kwargs=kwargs, field=field),
            "input_attributes": self._get_input_attribute(field=field, kwargs=kwargs),
            "field": field
        }

        if field.inputType == "checkbox" or field.inputType == "radio":
            params["label_class"] = PWebSSRConfig.INPUT_LABEL_CHECKBOX_CLASS_NAME
        else:
            params["label_class"] = PWebSSRConfig.INPUT_LABEL_CLASS_NAME

        params = self._get_select_options(field=field, params=params)
        return ssr_ui_render_html_file(self.html_file.form_input_html(), params=params)

    def get_input_error(self, field: FormField, kwargs):
        params = {
            "error_message_class": PWebSSRConfig.INPUT_ERROR_MESSAGE_CLASS_NAME,
            "field": field
        }
        return ssr_ui_render_html_file(self.html_file.error_message_html(), params=params)

    def get_input_help(self, field: FormField, kwargs):
        params = {
            "help_message_class": PWebSSRConfig.INPUT_HELP_MESSAGE_CLASS_NAME,
            "field": field
        }
        return ssr_ui_render_html_file(self.html_file.help_message_html(), params=params)

    def get_error_class(self, field: FormField, kwargs):
        if field.isError:
            return PWebSSRConfig.INPUT_ERROR_CLASS_NAME
        return ""

    def get_make_checked(self, key, value, kwargs):
        if str(key) == str(value):
            return "checked"
        return ""

    def get_make_select(self, key, value, kwargs):
        if str(key) == str(value):
            return "selected"
        return ""

    def get_set_value(self, value, kwargs):
        if value or value == 0:
            return f"value='{value}'"
        return value



from abc import ABC, abstractmethod
from jinja2 import nodes
from jinja2.ext import Extension

"""
Library Copied From: https://github.com/dldevinc/jinja2-simple-tags
"""


class BaseTemplateTag(Extension):

    def __init__(self, environment):
        super().__init__(environment)
        self.template = None
        self.lineno = None
        self.tag_name = None

    def parse(self, parser):
        lineno = parser.stream.current.lineno
        tag_name = parser.stream.current.value
        additional_params = [
            nodes.Keyword('_context', nodes.ContextReference()),
            nodes.Keyword('_template', nodes.Const(parser.name)),
            nodes.Keyword('_lineno', nodes.Const(lineno)),
            nodes.Keyword('_tag_name', nodes.Const(tag_name)),
        ]
        self.init_parser(parser)
        args, kwargs, target = self.parse_args(parser)
        kwargs.extend(additional_params)
        block_call = self.call_method('render_wrapper', args, kwargs)
        return self.output(parser, block_call, target, tag_name=tag_name, lineno=lineno)

    def init_parser(self, parser):
        parser.stream.skip(1)  # skip tag name

    def parse_args(self, parser):
        args = []
        kwargs = []
        require_comma = False
        target = None

        while parser.stream.current.type != 'block_end':
            if parser.stream.current.test('name:as'):
                parser.stream.skip(1)
                target = parser.stream.expect('name').value
                if parser.stream.current.type != 'block_end':
                    parser.fail(
                        'Invalid assignment target', parser.stream.current.lineno
                    )
                break

            if require_comma:
                parser.stream.expect('comma')

            if (
                    parser.stream.current.type == 'name'
                    and parser.stream.look().type == 'assign'
            ):
                key = parser.stream.current.value
                parser.stream.skip(2)
                value = parser.parse_expression()
                kwargs.append(nodes.Keyword(key, value, lineno=value.lineno))
            else:
                if kwargs:
                    parser.fail('Invalid argument syntax', parser.stream.current.lineno)
                args.append(parser.parse_expression())

            require_comma = True

        return args, kwargs, target

    def output(self, parser, block_call, target, tag_name, lineno):
        raise NotImplementedError

    def render_wrapper(self, *args, **kwargs):
        self.context = kwargs.pop('_context', None)
        self.template = kwargs.pop('_template', None)
        self.lineno = kwargs.pop('_lineno', None)
        self.tag_name = kwargs.pop('_tag_name', None)

        try:
            return self.render(*args, **kwargs)
        except BaseException as e:
            message = str(e).replace('render()', '').strip()
            raise Exception("TAG: {% " + self.tag_name + " %} " + message)

    @abstractmethod
    def render(self, *args, **kwargs):
        raise NotImplementedError


class StandaloneTag(BaseTemplateTag, ABC):
    def output(self, parser, block_call, target, tag_name, lineno):
        if target:
            target_node = nodes.Name(target, 'store', lineno=lineno)
            return nodes.Assign(target_node, block_call, lineno=lineno)
        call = nodes.MarkSafe(block_call, lineno=lineno)
        return nodes.Output([call], lineno=lineno)


class ContainerTag(BaseTemplateTag, ABC):
    def output(self, parser, block_call, target, tag_name, lineno):
        body = parser.parse_statements(['name:end%s' % tag_name], drop_needle=True)
        call_block = nodes.CallBlock(block_call, [], [], body).set_lineno(lineno)
        if target:
            target_node = nodes.Name(target, 'store', lineno=lineno)
            return nodes.AssignBlock(target_node, None, [call_block], lineno=lineno)
        return call_block



import pathlib
from ppy_file_text import FileUtil
from pweb_ssr.common.pweb_ssr_config import PWebSSRConfig


class PWebSSRHTMLFile:
    def get_html_path(self):
        if PWebSSRConfig.SSR_HTML_PATH:
            return PWebSSRConfig.SSR_HTML_PATH
        path = pathlib.Path(__file__).parent.parent.resolve()
        return FileUtil.join_path(path, "html")

    def form_input_html(self):
        return FileUtil.join_path(self.get_html_path(), "form-input.html")

    def error_message_html(self):
        return FileUtil.join_path(self.get_html_path(), "error-message.html")

    def help_message_html(self):
        return FileUtil.join_path(self.get_html_path(), "help-message.html")

    def pagination_html(self):
        return FileUtil.join_path(self.get_html_path(), "pagination.html")

    def sortable_header_html(self):
        return FileUtil.join_path(self.get_html_path(), "sortable-header.html")



from setuptools import setup, find_packages
import os
import pathlib

CURRENT_DIR = pathlib.Path(__file__).parent
README = (CURRENT_DIR / "README.md").read_text()

env = os.environ.get('source')


def get_dependencies():
    dependency = []

    if env and env == "code":
        return dependency

    return dependency + ["ppy-common", "pweb-form-rest"]


setup(
    name='pweb-ssr',
    version='0.0.4',
    url='https://github.com/banglafighter/pweb-ssr',
    license='Apache 2.0',
    author='Bangla Fighter',
    author_email='banglafighter.com@gmail.com',
    description='Server Side Rendered UI, Form, Table Header, Item Per Page, Pagination, Jinja Customization things',
    long_description=README,
    long_description_content_type='text/markdown',
    package_data={'pweb_ssr': ['html/**/*']},
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=get_dependencies(),
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
    ]
)
